/* 
 ===========================================================================
 =                                                                         =
 =  (C) Copyright 1991-1994 The Trustees of Indiana University             =
 =                                                                         =
 =  Permission to use, copy, modify, and distribute this program for       =
 =  non-commercial use and without fee is hereby granted, provided that    =
 =  this copyright and permission notice appear on all copies and          =
 =  supporting documentation, the name of Indiana University not be used   =
 =  in advertising or publicity pertaining to distribution of the program  =
 =  without specific prior permission, and notice be given in supporting   =
 =  documentation that copying and distribution is by permission of        =
 =  Indiana University.                                                    =
 =                                                                         =
 =  Indiana University makes no representations about the suitability of   =
 =  this software for any purpose. It is provided "as is" without express  =
 =  or implied warranty.                                                   =
 =                                                                         =
 ===========================================================================
 =                                                                         =
 = File:                                                                   =
 =   IUPOP3.C - Version 1.8                                                =
 =                                                                         =
 = Synopsis:                                                               =
 =   This file contains miscellaneous functions that support the           =
 =   IUPOP3 server for VMS.                                                =
 =                                                                         =
 = Authors:                                                                =
 =   Jacob Levanon & Larry Hughes                                          =
 =   Indiana University                                                    =
 =   University Computing Services, Network Applications                   =
 =                                                                         =
 = Credits:                                                                =
 =   This software is based on the Post Office Protocol version 3,         =
 =   as implemented by the University of California at Berkeley.           =
 =                                                                         =
 ===========================================================================
*/

/* ======================================================================== */
/* Includes */
/* ======================================================================== */

#if !defined(WINS) && !defined(MULTINET) && !defined(UCX)
    "Don't know how to make IUPOP3 for your TCP/IP implementation!"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef MULTINET
#include "multinet_root:[multinet.include]errno.h"
#else
#include <errno.h>
#endif /* MULTINET */

#include <types.h>
#include <time.h>
#include <in.h>
#include <socket.h>
#include <netdb.h>
#include <inet.h>
#include <unixio.h>

#include <starlet.h>       /* system services */
#include <iodef.h>
#include <ssdef.h>
#include <descrip.h>
#include <syidef.h>
#include <climsgdef.h>
#include <lib$routines>
#ifndef UCX
#include <vms/inetiodef.h>
#else
#include  <ucx$inetdef.h>      /* INET symbol definitions */
#endif

#ifdef KERBEROS
#include <krb.h>
#include <des.h>
#endif /* KERBEROS */

#include "iupop3_general.h"
#include "iupop3_vms.h"
#define Main
#include "iupop3_global.h"
#undef  Main
#include "version.h"

/* ======================================================================== */
/* Defines */
/* ======================================================================== */

#define ATTN_SLEEP      0
#define ATTN_CONNECT    1
#define ATTN_DISCONNECT 2
#define ATTN_TIMEOUT    3
#define ATTN_DATA       4
#define ATTN_UNBLOCK    5

#define LOGFILE_NEW   1
#define LOGFILE_FLUSH 2

#ifdef MULTINET
#define ERRNO socket_errno
#else
#define ERRNO errno
#endif

/* ======================================================================== */
/* Global Variables */
/* ======================================================================== */
int  pop_port   = RFC_PORT;
int initial_socket;
int connect_thread_number = 0; /* number of connecting thread */
int accept_has_failed = FALSE;

int client_timeout_interval[2]; /* used in issue_client_timeout_timer */
int blocked_interval[2]; /* delta time used in issue_client_unblock_timer */
int logfile_flush_interval[2]; /* delta time, maximum logfile flush interval */

POP pop[MAX_THREADS];
struct IOSB connect_iosb;

/* ======================================================================== */
/* Prototypes siehe iupop3_general.h */
/* ======================================================================== */

void  process_commandline_options (int argc, char *argv[]);
int init_global_timers (void);
int get_attn_state (int *attn_threadnum);
int has_expired (unsigned int time[2], unsigned int base_time[2]);

/* ======================================================================== */
/* Main */
/* ======================================================================== */
int main (int argc, char *argv[])
{
  state_table *s;
  POP *p;
  char *message;
  int  status;
  int  count;
  int  serving    = TRUE;
  int  attn_state = ATTN_SLEEP, attn_threadnum = 0;


  init_global_timers();
  process_commandline_options (argc, argv);
  get_date(start_time);
  get_time(start_time+11);
  start_time[10] = ' ';

  logfile_flush_ast ();

  system_log(LOG_INFO, "starting IUPOP3 V%s server on port %d",
             VERSION, pop_port);
  system_log(LOG_INFO, "client timeout value is %s (delta time)",
             time_from_bintime((unsigned int *)client_timeout_interval,0));
  system_log(LOG_INFO, "max new messages per client = %d", max_messages);

#ifdef KERBEROS
  if ((status = krb_get_lrealm(local_realm, 1)) != KSUCCESS)
  {
    system_log(LOG_ERROR, "realm lookup failed: %s", krb_err_txt[status]);
    exit(-1);
  }
  system_log(LOG_INFO, "local Kerberos realm is %s", local_realm);
#endif /* KERBEROS */

  gethostname(myhostname, MAXHOSTNAMELEN);
  lower (myhostname);
  if (!create_initial_socket(pop_port))
  {
    system_log(LOG_ERROR, "could not create or bind initial socket");
    exit (EXIT_FAILURE);
  }

  for (count=0; count<MAX_THREADS; count++)
     pop[count].in_use = FALSE;

  listen(initial_socket, SOMAXCONN);
  issue_new_connect_qio();
  attn_state = ATTN_SLEEP;

  /* run loop */
  while (serving)
  {
    if ((current_threads == 0) && server_shutdown) break;
    attn_state = get_attn_state (&attn_threadnum);
    main_loop_count++;
    p = &pop[attn_threadnum];
    switch (attn_state)
    {
      case ATTN_CONNECT:
        total_threads++;
        if (process_new_connect(attn_threadnum))
        {
#ifdef KERBEROS
          if (client_authenticated(p))
          {
            issue_client_read_qio(attn_threadnum);
            issue_client_timeout_timer(attn_threadnum);
          }
          else
          {
            authentication_failures++;
            close_pop_thread(p);
          }
#else /* not KERBEROS */
          issue_client_read_qio(attn_threadnum);
          issue_client_timeout_timer(attn_threadnum);
#endif /* KERBEROS */
        }
        else
        {
          pop_log(LOG_ERROR, p, "new connect failed");
          close_pop_thread(p);
        }
        issue_new_connect_qio();
        break;

      case ATTN_DISCONNECT:
        abnormal_disconnects++;
        pop_log(LOG_ERROR, p, "abnormal disconnect");
        cancel_client_timer(attn_threadnum);
        close_pop_thread(p);
        break;

      case ATTN_TIMEOUT:
        timeouts++;
        pop_log(LOG_ERROR, p,"timed out (originally %s)",
                           time_from_bintime(pop[attn_threadnum].timeout,1));
        close_pop_thread(p);
        break;

      case ATTN_DATA:
        cancel_client_timer(attn_threadnum);
        process_thread(p);
        if (pop[attn_threadnum].CurrentState == halt)
        {
          normal_disconnects++;
          pop_log(LOG_DEBUG, p, "normal disconnect");
          close_pop_thread(p);
        }
        else if (pop[attn_threadnum].authentication_attempts >=
                 MAX_AUTHENTICATION_ATTEMPTS)
        {
          pop_log(LOG_ERROR, p, "auth attempts exceeded");
          close_pop_thread(p);
        }
        else if (pop[attn_threadnum].retrieve.blocked)
        {
          issue_client_unblock_timer(attn_threadnum);
        }
        else 
        {
          if (strstr (p->command_buffer, "\r\n"))
            p->attn_state = ATTN_DATA;
          else
          {
            issue_client_read_qio(attn_threadnum);
            issue_client_timeout_timer(attn_threadnum);
          }
        }
        break;

      case ATTN_UNBLOCK:
        if (pop[attn_threadnum].retrieve.tries++ <= 12)
        {
          mail_retrieve_message_text(p);
          if (pop[attn_threadnum].retrieve.blocked)
            issue_client_unblock_timer(attn_threadnum);
          else
          {
            issue_client_read_qio(attn_threadnum);
            issue_client_timeout_timer(attn_threadnum);
          }
        }
        else
        {
          pop_log(LOG_ERROR, p, 
                  "socket blocked for too long...disconnecting");
          close_pop_thread(p);
        }
        break;
      case ATTN_SLEEP:
        hibernation_count++;
        status = sys$hiber();
        break;
      default:
        system_log(LOG_ERROR, "invalid attn_state: %d, attn_threadnum = %d",
                   attn_state, attn_threadnum);
        serving = FALSE;
        break;
    }
  }

  system_log(LOG_INFO, "server is shutting down");
  fclose(log_file);

#ifdef UCX
  status = sys$cancel((unsigned)vaxc$get_sdc(initial_socket));
#else
  status = sys$cancel(initial_socket);
#endif

  netclose(initial_socket);

  exit (EXIT_SUCCESS);
}


/* ======================================================================== */
/* Process the commandline parameters */
/* ======================================================================== */
void  process_commandline_options (int argc, char *argv[])
{
  int  bad_option = FALSE;
  int  status;

  while ((argc > 1) && !bad_option)
  {
    argc--; argv++;

    if (strcmp(argv[0], "-port") == 0)
    {
      argc--; argv++;
      if (!argc)
        bad_option = TRUE;
      else
      {
        pop_port = atoi(argv[0]);
        if (pop_port <= 0) bad_option = TRUE;
      }
    }

    else if (strcmp(argv[0], "-loglevel") == 0)
    {
      argc--; argv++;
      if (!argc)
        bad_option = TRUE;
      else
      {
        if (strcmp(argv[0], "debug") == 0)
          master_log_level = LOG_DEBUG;
        else if (strcmp(argv[0], "thread") == 0)
          master_log_level = LOG_THREAD;
        else if (strcmp(argv[0], "info") == 0)
          master_log_level = LOG_INFO;
        else if (strcmp(argv[0], "error") == 0)
          master_log_level = LOG_ERROR;
        else
          bad_option = TRUE;
      }
    } 

    else if (strcmp(argv[0], "-logfile") == 0)
    {
      argc--; argv++;
      if (!argc)
        bad_option = TRUE;
      else
        log_filename = argv[0];
    }

    else if (strcmp(argv[0], "-maxmsg") == 0)
    {
      argc--; argv++;
      if (!argc)
        bad_option = TRUE;
      else
      {
        max_messages = atoi(argv[0]);
        if (max_messages <= 0) max_messages = MAX_MESSAGES;
      }
    }

    else if (strcmp(argv[0], "-purge_reclaim") == 0)
    {
      purge_reclaim = TRUE;
    }

    else if (strcmp(argv[0], "-xtnd") == 0)
    {
      argc--; argv++;
      if (!argc)
        bad_option = TRUE;
      else
        xtnd_filename = argv[0];
    }

    else if (strcmp(argv[0], "-timeout") == 0)
    {
      argc--; argv++;
      if (!argc)
        bad_option = TRUE;
      else
      {
        int interval = atoi (argv[0]);
        if ((interval < 1) || (interval > (24*60-1)))
        {
           bad_option = TRUE;
           printf ("timeout value out of range, valid are 1 to %d minutes\n",(24*60-1));
        }
        else
        {
           int hours, minutes;
	   char time_str[sizeof "0 00:02:00.00"];
           hours = interval / 60;
           minutes = interval % 60;
           sprintf (time_str, "0 %02.2d:%02.2d:00.00", hours, minutes);
	   status = sys$bintim(Descr(time_str),client_timeout_interval);
	   if (vms_error(status))
           {
              fprintf (stderr,"can't change client timer %s", 
                               vms_message(status));
              bad_option = TRUE;
           }
        }
      }
    }
    else if (strcmp(argv[0], "-default_type") == 0)
    {
      argc--; argv++;
      if (!argc)
        bad_option = TRUE;
      else
      {
        lower(argv[0]);
        if (strstr (argv[0], "smtp"))
          default_message_type = SMTP_MESSAGE;
        else if (strstr (argv[0], "internet"))
          default_message_type = SMTP_MESSAGE;
        else if (strstr (argv[0], "decnet"))
          default_message_type = DECNET_MESSAGE;
      }
    }

    else
      bad_option = TRUE;
  }

  if (bad_option)
  {
    printf("Usage: iupop3 [-port n] [xtnd file] [-loglevel level]\n");
    printf("              [-logfile file] [-maxmsg N] [-purge_reclaim]\n");
    printf("              [-default_type type] [-timeout N] \n");
    exit (CLI$_INVQUAL);
  }
  return;
}

/* ======================================================================== */
/* Cancel Client Timer */
/* ======================================================================== */
int cancel_client_timer (int threadnum)
{
  int status = TRUE;
  POP * p = &pop[threadnum];

  p->timeout[0] = p->timeout[1] = 0;
  return status;
}

/* ======================================================================== */
/* Close POP Thread */
/* ======================================================================== */
int close_pop_thread (POP *p)
{
  static int num_too_many_threads = 0;
  int  status = SS$_NORMAL;

  if (p->in_use)
  {
    current_threads--;
    pop_log (LOG_THREAD, p, "closing thread");
    if (p->channel) status = sys$cancel(p->channel);
    if (vms_error(status)) 
      pop_log(LOG_ERROR, p,"sys$cancel: %s", vms_message(status));
    p->retrieve.bufsize = 10;
#ifdef UCX
    if (p->sockfd)
    {
      status = getsockopt (p->sockfd, SOL_SOCKET, SO_ERROR,
                 p->retrieve.buffer,(void *) &p->retrieve.bufsize);
      if (status == -1) 
         pop_log(LOG_ERROR,p,"getsockopt: status %d, size %d, errno %d",
                               status,p->retrieve.bufsize,ERRNO);
    }
#endif
    if (p->sockfd) status = netclose (p->sockfd);
    if (status == -1)
      pop_log (LOG_ERROR, p,"netclose: error closing socket %d, errno %d",
      p->sockfd, ERRNO);

    if ((int) p->message_context != NO_CONTEXT)
      mail_close_message_context(p);
    if ((int) p->file_context != NO_CONTEXT)
      mail_close_file_context(p);
    if ((int) p->user_context != NO_CONTEXT)
      mail_close_user_context(p);

    if (p->retrieve.ext_file) fclose (p->retrieve.ext_file);
    if (p->retrieve.send_buffer) free (p->retrieve.send_buffer);
    if (p->mptr) free (p->mptr);
    memset(p,0,sizeof(pop[0]));
    p->in_use = FALSE;
    status = TRUE;
    if (too_many_threads > num_too_many_threads)
    {
      num_too_many_threads = too_many_threads;
      system_log (LOG_THREAD, "reinitiate new_connect_ast");
      new_connect_ast ();
    }
  }
  else
  {
    pop_log(LOG_ERROR, p, "thread not in use -- cannot close it!");
    status = FALSE;
  }
  return status;
}

/* ======================================================================== */
/* Client Read AST */
/* ======================================================================== */
void client_read_ast (int threadnum)
{
  int status;
  POP * p = &pop[threadnum];

  status = p->read_iosb.status;
  if (p->attn_state != ATTN_SLEEP)
    pop_log (LOG_ERROR, p, "client_read_ast: attn_state is %d", p->attn_state);
  if (!p->in_use || (status == SS$_CANCEL) || (status == SS$_ABORT) )
    p->attn_state = ATTN_TIMEOUT;
  else
  {
    if (vms_error(p->read_iosb.status))
      pop_log(LOG_ERROR, p,"read iosb: %s",
              vms_message(pop[threadnum].read_iosb.status));

    if (vms_error(p->read_iosb.status) ||
       (p->read_iosb.terminator_offset <= 0))
    {
      /* Be paranoid and flag it as disconnected right now */
      p->attn_state = ATTN_DISCONNECT;
      p->connected = FALSE;
    }
    else
      p->attn_state = ATTN_DATA;
  }
  /* attn_threadnum = threadnum; */
  status = sys$wake(0,0);
  if (vms_error(status))
    pop_log(LOG_ERROR, p,"sys$wake: %s", vms_message(status));
}

/* ======================================================================== */
/* Client Timeout AST */
/* ======================================================================== */
void client_timeout_ast (int threadnum)
{
  int status;

  /*
   * Just cancel, and let the read ast handle it.  Remember
   * that the threadnum parameter needs to be decremented
   * since it was incremented when calling sys$setimr.
   */
  status = sys$cancel(pop[--threadnum].channel);
}

/* ======================================================================== */
/* Client Unblock AST */
/* ======================================================================== */
void client_unblock_ast (int threadnum)
{
  int status;
  /*
   * Remember that the threadnum parameter needs to be decremented
   * since it was incremented when calling sys$setimr.
   */
  POP *p = &pop[--threadnum];

  p->attn_state = ATTN_UNBLOCK;
  status = sys$wake(0,0);
}

/* ======================================================================== */
/* Create Initial Socket */
/* ======================================================================== */
int create_initial_socket(int pop_port)
{
  int return_value = FALSE;
  int status;
  struct sockaddr_in server;
  static int one = 1;

  initial_socket = socket(AF_INET,SOCK_STREAM,0);
  if (initial_socket < 0)
    system_log(LOG_ERROR, "error creating initial socket");
  else
  {
    if (setsockopt(initial_socket, SOL_SOCKET, SO_KEEPALIVE, 
                        (char *)&one, sizeof(one)) == -1)
      system_log(LOG_ERROR, "error setting SO_KEEPALIVE");

    if (setsockopt(initial_socket, SOL_SOCKET, SO_REUSEADDR,
                        (char *)&one, sizeof(one)) == -1)
      system_log(LOG_ERROR, "error setting SO_REUSEADDR");

    server.sin_family      = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port        = htons (pop_port);

/*
 * The 2nd argument of bind should be a structure from type "sockaddr",
 * but in UCX$EXAMPLES they use struct "sockaddr_in", too.
 */
    status = bind (initial_socket, (struct sockaddr *)&server, sizeof(server));
    if (status != 0)
      system_log(LOG_ERROR, "error binding initial socket");
    else
      return_value = TRUE;
  }
  return(return_value);
}

/* ====================================================================== */
/* Init POP Thread */
/* ====================================================================== */
int init_pop_thread (int threadnum, POP *p, int sockfd)
{
    int status;
    size_t len;
    char *pointer;
    struct hostent *ch;

    system_log(LOG_THREAD, "thread %2d: initializing thread", threadnum);

    /* Initialize this POP structure -- all fields below are vital! */
    p->sockfd            = sockfd;
#ifdef UCX
    p->channel		 = vaxc$get_sdc(p->sockfd);
#else
    p->channel		 = p->sockfd;
#endif
    p->threadnum         = threadnum;
    p->bytes_deleted     = 0;
    p->msg_count         = 0;
    p->newmail_size      = 0;
    p->msgs_deleted      = 0;
    p->newmail_selected  = FALSE;
    p->user_context      = NO_CONTEXT;
    p->file_context      = NO_CONTEXT;
    p->message_context   = NO_CONTEXT;
    p->CurrentState      = auth1;
    p->connected         = TRUE;
    p->retrieve.send_ptr = p->retrieve.send_buffer;

    memset(p->read_buffer, '\0', READ_BUFFERSIZE);
    memset(p->command_buffer, '\0', READ_BUFFERSIZE);
    p->command_pointer = p->command_buffer;

    /*
    **  Get the address and socket of the client to whom we're speaking
    */
    len = sizeof(struct sockaddr_in);
    if ((status = getpeername
                  (sockfd,(struct sockaddr *)&p->sin,(void *) &len)) < 0)
    {
      pop_log(LOG_ERROR, p, "getpeername failed: error %d", status);
      return(FALSE);
    }
    pointer = inet_ntoa(p->sin.sin_addr);
    strcpy(p->ipaddr, pointer);
    p->client = p->ipaddr;
    p->ipport = ntohs(p->sin.sin_port);
    system_log(LOG_THREAD, "thread %2d: client address is %s,%d", p->threadnum,
               p->client, p->ipport);

    /*
    **  Everything succeeded, so mark this thread as "in use"
    */
    p->in_use = TRUE;
    return TRUE;
}

#ifdef KERBEROS
/* ======================================================================== */
/* Client Authenticated */
/* ======================================================================== */
int client_authenticated(POP *p)
{
  int retval = FALSE;
  int status;

  /* Receive the client's Kerberos ticket */
  strcpy(p->instance, "*");
  status = krb_recvauth(0L, p->sockfd, &p->ticket, "pop",
                        p->instance, &p->sin,
                        (struct sockaddr_in *) NULL, &p->kdata, "",
                        p->schedule, p->version);
  if (status != KSUCCESS)
  {
    pop_log(LOG_ERROR, p, "ticket receive error: %s", krb_err_txt[status]);
    pop_msg(p, POP_FAILURE, "ticket receive error: %s", krb_err_txt[status]);
  }
  else
  {
    /* Partial success -- ticket at least is OK */
    strcpy(p->user, p->kdata.pname);
    pop_log(LOG_THREAD, p, "ticket: %s.%s@%s (%s)", p->kdata.pname,
               p->kdata.pinst, p->kdata.prealm,
               inet_ntoa(p->sin.sin_addr));

    /* Is the user in OUR Kerberos realm? */
    if (strcmp(p->kdata.prealm, local_realm) != 0)
    {
      pop_log(LOG_ERROR, p, "(%s.%s@%s) realm not accepted", p->kdata.pname,
                 p->kdata.pinst, p->kdata.prealm);
      pop_msg(p, POP_FAILURE, "(%s.%s@%s) realm not accepted", p->kdata.pname,
                 p->kdata.pinst, p->kdata.prealm);
    }
    else
    {
      /* Usernames should have a null instance! */
      if (strcmp(p->kdata.pinst, ""))
      {
        pop_log(LOG_ERROR, p, "(%s.%s@%s) instance not accepted",
                   p->kdata.pname, p->kdata.pinst,
                   p->kdata.prealm);
        pop_msg(p, POP_FAILURE, "(%s.%s@%s) instance not accepted",
                   p->kdata.pname, p->kdata.pinst,
                   p->kdata.prealm);
      }
      else
      {
        pop_log(LOG_DEBUG, p, "Kerberos authenticated - client is '%s'", 
                p->user);
        if (valid_vms_user(p))
        {
          pop_log(LOG_DEBUG, p, "vms authenticated");
          pop_msg(p, POP_SUCCESS, "IUPOP3 server (Kerberos) V%s at %s, up since %s",
                  VERSION, myhostname, start_time);
          retval = TRUE;
        }
        else
          pop_msg(p, POP_FAILURE, "user account \"%s\" not available", p->user);
      }
    }
  }
  return(retval);
}
#endif /* KERBEROS */

/* ======================================================================== */
/* Issue Client Read QIO */
/* ======================================================================== */
int issue_client_read_qio(int threadnum)
{
  int status = SS$_NORMAL;
  POP *p = &pop[threadnum];

  if (p->connected)
  {
    memset(p->read_buffer, '\0', READ_BUFFERSIZE);
    status = sys$qio(0, pop[threadnum].channel, IO$_READVBLK,
                     &pop[threadnum].read_iosb, client_read_ast, threadnum, 
                     pop[threadnum].read_buffer, READ_BUFFERSIZE, 0, 0, 0, 0);
    if (vms_error(status))
      pop_log(LOG_ERROR, &pop[threadnum], "client read sys$qio: %s", 
              vms_message(status));
  }
  else
  {
    p->attn_state = ATTN_DISCONNECT;
  }
  return status;
}

/* ======================================================================== */
/* Issue Client Timeout Timer */
/* ======================================================================== */
int issue_client_timeout_timer (int threadnum)
{
  int status;
  unsigned int current_time[2];
  POP * p = &pop[threadnum];

  sys$gettim (current_time);
  status = lib$add_times (client_timeout_interval, current_time, p->timeout);
  return status;
}

/* ======================================================================== */
/* Issue Client Unblock Timer */
/* ======================================================================== */
int issue_client_unblock_timer(int threadnum)
{
  int status = SS$_NORMAL;
  int value = FALSE;
  int interval[2];
  int mult = pop[threadnum].retrieve.tries;

  mult *= 5;
  interval[0] = blocked_interval[0]; interval[1] = blocked_interval[1];
  if (mult) status = lib$mult_delta_time (&mult,interval);
  if (vms_error(status))
  {
    pop_log(LOG_ERROR, &pop[threadnum], "lib$mult_delta_time (mult %d): %s",
                                         mult, vms_message(status));
    interval[0] = blocked_interval[0]; interval[1] = blocked_interval[1];
  }
  /* Use "threadnum+1" since sys$cantim uses 0 to cancel all requests */
  status = sys$setimr(0,interval,client_unblock_ast,threadnum+1,0);
  if (vms_error(status))
  {
    pop_log(LOG_ERROR, &pop[threadnum], "sys$setimr (unblock): %s", 
                                         vms_message(status));
    client_unblock_ast (threadnum+1);
  }
  else
    value = TRUE;
  return value;
}

/* ======================================================================== */
/* Issue New Connect QIO */
/* ======================================================================== */
int issue_new_connect_qio (void)
{
  int status;

#ifdef UCX
  status = sys$qio(0, (unsigned)vaxc$get_sdc(initial_socket), 
                  IO$_SETMODE | IO$M_READATTN, &connect_iosb, 0, 0,
		   &new_connect_ast, 0, 0, 0, 0, 0);
#else
  status = sys$qio(0, (u_long)initial_socket, IO$_ACCEPT_WAIT, 
                   &connect_iosb, &new_connect_ast, 0, 0, 0, 0, 0, 0, 0);
#endif
  if (!vms_error(status))
    status = TRUE;
  else
  {
    system_log(LOG_ERROR, "new connect sys$qio: %s", vms_message(status));
    status = FALSE;
  }
  return status;
}

/* ======================================================================== */
/* New Connect AST */
/* ======================================================================== */
void new_connect_ast (void)
{
  static int max_threads = MAX_THREADS;
  int threadnum, status;
  POP *p = &pop[0];

  if (accept_has_failed)  /* last accept() has failed */
  {
    /* if the last accept has failed because of insufficient quotas or 
       recources, reducing the maximum number of threads probably helps */
    if (current_threads) 
    {
      too_many_threads++;
      return; /* restart this routine after next thread is closed */
    }
    else
    {
      accept_has_failed = FALSE;
      if (initial_socket)
      {
        status = netclose (initial_socket);
        if (status == -1)
        {
          system_log(LOG_ERROR, "could not close initial socket");
          exit (EXIT_FAILURE);
        }
        else
        {
          system_log(LOG_INFO, "create initial socket");
          if (!create_initial_socket(pop_port))
          {
             system_log(LOG_ERROR, "could not create or bind initial socket");
             exit (EXIT_FAILURE);
          }
          listen(initial_socket, SOMAXCONN);
        }
      }
      if (max_threads > 5)
      {
        max_threads -= 5;
        system_log (LOG_INFO, "reducing maximum threads to %d",max_threads);
      }
      else
        exit (EXIT_FAILURE);  /* the server doesn't work! */
      issue_new_connect_qio();
    }
  }

  if (vms_error(connect_iosb.status))
    system_log(LOG_ERROR, "connect iosb: %s",vms_message(connect_iosb.status));
  for (threadnum=0; (threadnum < max_threads) && p->in_use; threadnum++,p++);
  if (threadnum < MAX_THREADS)
  {
    current_threads++;
    p->in_use = TRUE;
    p->threadnum = threadnum;
    p->attn_state = ATTN_CONNECT;
    sys$gettim (p->start_time);
    connect_thread_number = threadnum + 1;
  }
  else
  {
    /* leave without doing anything. close_pop_thread () will rerun
       this routine if a thread becomes available. */
    too_many_threads++;
    system_log (LOG_THREAD, "new_connect_ast warning: too many threads!");
  }
  status = sys$wake(0,0);
}

/* ======================================================================== */
/* Open or Flush Log File */
/* ======================================================================== */
int open_log_file (int modus)
{
  int retval;
  int status;

  if (log_filename == NULL)
  { 
    log_file = stdout;
    retval = FALSE;
  }
  else
  {
    retval = TRUE;
    if (modus == LOGFILE_NEW)
      log_file = fopen (log_filename, "w", "shr=upd", "dna=.log", "rfm=var");
    else if (modus == LOGFILE_FLUSH)
    {
      fflush (log_file);
      status = fsync ( fileno(log_file));
      if (status == -1)
        system_log(LOG_ERROR, "fsync: errno %d", ERRNO);
    }
    else
    {
      fprintf (stderr, "open_log_file: unknown mode %d",modus);
      if (log_file) fclose (log_file);
      log_file = stdout;
      retval = FALSE;
    }
    if (!log_file)
    {
      log_file = stdout;
      system_log(LOG_ERROR, "log file could not be opened: error %d", ERRNO);
      retval = FALSE;
    }
  }
  return retval;
}

/* ======================================================================== */
/* Issue Logfile Flush Timer */
/* ======================================================================== */
int issue_logfile_flush_timer (void)
{
  static int first = TRUE, timer_efn = 0, interval[2];
  int status = SS$_NORMAL;
  int value = FALSE;
  unsigned int efncluster;

  if (first)
  {
    first = FALSE;
    status = sys$bintim(Descr ("0 00:00:30.00"), interval);
    if (vms_error(status))
      system_log (LOG_ERROR, "sys$bintim: %s", vms_message(status));
    status = lib$get_ef (&timer_efn);
    if (vms_error(status))
      system_log (LOG_ERROR, "lib$get_ef: %s", vms_message(status));
    sys$setef (timer_efn);
  }

  if (sys$readef(timer_efn, &efncluster) == SS$_WASSET)
  {
    status = sys$setimr(timer_efn, interval, logfile_flush_ast, 0, 0);
    if (vms_error(status))
    {
      sys$setef (timer_efn);
      system_log (LOG_ERROR, "sys$setimr (logfile): %s", vms_message(status));
    }
    else
      value = TRUE;
    system_log (LOG_DEBUG, "issue logfile flush timer");
  }

  return value;
}

/* ======================================================================== */
/* Logfile Flush AST */
/* ======================================================================== */
void logfile_flush_ast (void)
{
  static int day = 0;
  unsigned int current_time[2];
  short int numtime[7];

  sys$gettim (current_time);
  sys$numtim (numtime, current_time);

  if (day != numtime[2])
  {
    open_log_file (LOGFILE_NEW);
    if (day)
      system_log (LOG_INFO, "new log file, up since %20.20s", 
                            time_from_bintime (current_time,0));
  }
  else
  {
    open_log_file (LOGFILE_FLUSH);
  }
  day = numtime[2];
}

/* ======================================================================== */
/* Process New Connect */
/* ======================================================================== */
int process_new_connect (int threadnum)
{
  static int can_set_sndbuf = TRUE;
  int  retval = FALSE;
  int  ok = FALSE;
  int  status;
  int  new_socket;
  size_t  length;
  POP  *p;
  
  struct sockaddr addr;
  static int one = 1;
  static int sockbuf = SOCK_BUFSIZE;

  /*
  **  on different TCIP/IP stacks and VMS versions the third argument
  **  of accept is sometimes defined as int * or insigned int *, so
  **  the value is passed as void *
  */
  length = sizeof(addr);
  new_socket = accept (initial_socket,&addr,(void *)&length);
  if (new_socket < 0)
  {
    system_log(LOG_ERROR, "accept failed, error number %d", ERRNO);
    accept_has_failed = ERRNO;
  }
  else
  {
    if (threadnum >= 0)
    {
      if (setsockopt(new_socket, SOL_SOCKET, SO_KEEPALIVE,
                 (char *)&one, sizeof(one)) == -1)
        system_log(LOG_ERROR, "error setting SO_KEEPALIVE");

      if (can_set_sndbuf)
      {
        if (setsockopt(new_socket, SOL_SOCKET, SO_SNDBUF,
                   (char *)&sockbuf, sizeof(sockbuf)) == -1)
        {
          can_set_sndbuf = FALSE;
          system_log(LOG_ERROR, "error setting SO_SNDBUF");
        }
      }
      p = &pop[threadnum];
      if (current_threads > maximum_threads) maximum_threads = current_threads;
      system_log(LOG_DEBUG, "new connection accepted: assigned thread %d",
                 threadnum);
      if (init_pop_thread(threadnum,p,new_socket))
      {
#ifndef KERBEROS
        char timestamp[80];
	sprintf (timestamp, TIMESTAMP_FORMAT,
	         VERSION, myhostname, time_from_bintime (p->start_time, 0));
        pop_msg (p, POP_SUCCESS, "IUPOP3 server %s",timestamp);
#endif /* !KERBEROS */
        retval = TRUE;
        /* attn_threadnum = threadnum; */ /* assign for later qio issue */
      }
      else
      {
        system_log(LOG_ERROR, "thread %d could not be initialized; closing",
                   threadnum);
        netclose(new_socket);
      }
    }

#if 0
    else
    {
      char message[80];
      sprintf(message, "%s too many threads!\r\n", POP_ERR);
#ifdef UCX
      sys$qio(0, (unsigned)vaxc$get_sdc(new_socket), IO$_WRITEVBLK,
                   0, 0, 0, message, strlen(message), 0, 0, 0, 0);
#else
      sys$qio(0, (unsigned)new_socket, IO$_WRITEVBLK,
                   0, 0, 0, message, strlen(message), 0, 0, 0, 0);
#endif
      too_many_threads++;
      system_log(LOG_ERROR, "error: too many threads!");
      netclose(new_socket);
    }
#endif

  }
  return retval;
}

/* ======================================================================== */
/* Process Thread */
/* ======================================================================== */
void process_thread (POP *p)
{
  int status, size;
  state_table *s;
  char *pointer, *tmp;

  /* Check to see if single character == backspace or delete */
  size = p->read_iosb.terminator_offset;
  if (size)
  {
    net_bytes_read += size;
    net_read_count++;
    if ((size == 1) && 
        ((*p->read_buffer == 8) || (*p->read_buffer == 127)))
    {
      if (p->command_pointer > p->command_buffer)
        *(--p->command_pointer) = '\0';
      return;
    }
    /* Move data from read_buffer to latest position in command_buffer */
    memcpy(p->command_pointer, p->read_buffer, size);
    p->command_pointer += size;
    *p->command_pointer = '\0';
    p->read_iosb.terminator_offset = 0;
  }

  /* Check for the CR/LF terminator; if it exists, continue. */
  if ((pointer = strstr (p->command_buffer, "\r\n")))
  {
    /* Remove the terminator, and advance the pointer. */
    *pointer = '\0';
    pointer+=2;
    if (strlen(p->command_buffer) > 0)
    {
      commands_processed++;
      pop_log(LOG_THREAD, p, "rx: %s", 
        (strncasecmp(p->command_buffer, "pass ", 5) ? 
        p->command_buffer : "pass"));
       /*
       * If the previous command was "retr", it is finally safe to
       * mark the message as successfully retrieved.
       */
      if (p->previous_command)
      {
        if ((strcmp(p->previous_command, "retr") == 0) && 
            p->retrieve.completed)
        {
          Message *mp;
          if (p->retrieve.message_id > p->last_msg)
            p->last_msg = p->retrieve.message_id;
          mp = &(p->mptr[p->last_msg-1]);
          pop_log(LOG_DEBUG, p, "marking message #%d as retrieved", mp->number);
          mp->retr_flag = TRUE;
        } 
      } 
       /* Process the command. */
      if ((s = pop_get_command(p, p->command_buffer)) != NULL)
      {
        if (s->function)
          p->CurrentState = s->PostState[(*s->function)(p)];
        else
        {
          p->CurrentState = s->PostState[0];
          pop_msg(p, POP_SUCCESS, NULL);
        }
      }
       /* Retain previous command */
      if (strlen(p->pop_args[0]) < sizeof(p->previous_command))
        strcpy(p->previous_command, p->pop_args[0]);
      else
      {
        strncpy(p->previous_command, p->pop_args[0],
          sizeof(p->previous_command)-1);
        p->previous_command[sizeof(p->previous_command)-1] = '\0';
      }
    }
    /* Slide the buffer back to the beginning. */
    for (tmp = p->command_buffer;pointer < p->command_pointer; pointer++)
      *tmp++ = *pointer;
    p->command_pointer = tmp;
    *p->command_pointer = '\0';
  }
  sys$wake (0,0);
}

/* ======================================================================== */
/* init global timer values used in issue_client_timeout_timer                           */
/* and issue_client_unblock_timer                                           */
/* ======================================================================== */
int init_global_timers (void)
{
  int status;
  int interval[2];

  status = sys$bintim (Descr("0 00:00:30.00"),interval);
  status = sys$schdwk (0,0,interval,interval);
  if (vms_error(status))
    fprintf (stderr, "sys$schdwk: %s", vms_message(status));

  status = sys$bintim(Descr("0 00:02:00.00"),client_timeout_interval);
  if (vms_error(status))
    fprintf (stderr, "sys$bintim: %s", vms_message(status));

  status = sys$bintim(Descr("0 00:00:00.10"),blocked_interval);
  if (vms_error(status))
    fprintf (stderr, "sys$bintim: %s", vms_message(status));

  status = sys$bintim(Descr("0 00:05:00.00"),logfile_flush_interval);
  if (vms_error(status))
    fprintf (stderr, "sys$bintim: %s", vms_message(status));
  return status;
}


/* ======================================================================== */
/* check threads for attn_state */
/* ======================================================================== */
int get_attn_state (int *attn_threadnum)
{
  static int num_att=0, states[MAX_THREADS], threadnum[MAX_THREADS];
  static unsigned int logfile_flush_timer[2];
  int attn_state = ATTN_SLEEP;
  POP *p;

  /* process new connect first */
  if (connect_thread_number)
  {
     p = &pop[--connect_thread_number];
     attn_state = p->attn_state;
     *attn_threadnum = connect_thread_number;
     p->attn_state = ATTN_SLEEP;
     connect_thread_number = 0;
     return attn_state;
  }

  if (!num_att)
  {
    unsigned int current_time[2];
    int num_threads = current_threads;
    int timeout;
    int i;

    sys$gettim (current_time);
    for (i = 0, p = &pop[0]; num_threads && (i < MAX_THREADS); i++,p++)
    {
      if (p->in_use)
      {
        num_threads--;
        if (p->attn_state != ATTN_SLEEP)
        {
          states[num_att] = p->attn_state;
          threadnum[num_att++] = i;
          p->attn_state = ATTN_SLEEP;
        }
        /* check for client timeouts */
        else if (p->timeout[0] || p->timeout[1])
        {
           if (has_expired (p->timeout, current_time))
             client_timeout_ast (i+1);
        }
      }
    }
    if (has_expired (logfile_flush_timer, current_time))
    {
      issue_logfile_flush_timer ();
      lib$add_times (logfile_flush_interval, current_time, logfile_flush_timer);
    }
  }

  if (num_att)
  {
    attn_state = states[--num_att];
    *attn_threadnum = threadnum[num_att];
  }
  return attn_state;
}

/* ======================================================================== */
/* checks two absolute times, returns TRUE if time lower than base_time    */
/* ======================================================================== */
int has_expired (unsigned int time[2], unsigned int base_time[2])
{
  int expired;

  if (base_time[1] < time[1])
    expired = FALSE;
  else if (base_time[1] > time[1])
    expired = TRUE;
  else if (base_time[0] > time[0])
    expired = TRUE;
  else
    expired = FALSE;
  return expired;
}
